@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 2008-2016  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 2008-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -89,6 +89,15 @@ static ssize_t _connchain_recv (struct connchain_i **chain, idx_t idx,
   return i;
 }
 
+/* callback from sockets... be sure peer->iface is alive! */
+static void _connchain_do_mark (void *data)
+{
+  struct peer_t *peer = data;
+
+  if (peer->iface)
+    Mark_Iface (peer->iface);
+}
+
 /* allocate starting link structure */
 static void _connchain_create (struct peer_t *peer)
 {
@@ -101,6 +110,7 @@ static void _connchain_create (struct peer_t *peer)
   peer->connchain->next = NULL;
   peer->connchain->tc = 0;
   peer->connchain->buf = (struct connchain_buffer *)peer;
+  AssociateSocket(peer->socket, &_connchain_do_mark, peer);
   dprint (2, "connchain.c: created initial link %p", peer->connchain);
 }
 
@@ -236,6 +246,9 @@ ssize_t Connchain_Get (connchain_i **chain, idx_t idx, char *buf, size_t sz)
     return i;
   if(Connchain_Get (&(*chain)->next, idx, NULL, 0))i=i; /* it's dead, kill next */
   dprint (2, "connchain.c: destroying link %p", *chain);
+  /* connchain is dead, disassociate it from the socket ASAP */
+  if ((*chain)->tc == 0)
+    AssociateSocket(idx, NULL, NULL);
   pthread_mutex_lock(&CC_Mutex);
   free_connchain_i (*chain);
   pthread_mutex_unlock(&CC_Mutex);
@@ -251,7 +264,7 @@ typedef struct				/* one side buffer for filter 'x' */
 {
   ssize_t inbuf;			/* how much bytes are in buf */
   size_t bufpos;			/* where is first char */
-  char buf[2*MB_LEN_MAX*MESSAGEMAX];	/* message buffer */
+  char buf[2*MBMESSAGEMAX];		/* message buffer */
 } connchain_b;
 
 struct connchain_buffer			/* local buffer type for filter 'x' */
@@ -269,10 +282,12 @@ static ssize_t _ccfilter_x_push (connchain_i **ch, idx_t id, connchain_b *bb)
 
   DBG("connchain.c:_ccfilter_x_send: sending buffer =%zu +%zd", bb->bufpos, bb->inbuf);
   i = Connchain_Put (ch, id, &bb->buf[bb->bufpos], &bb->inbuf);
-  if (i < 0)				/* some error, end us */
+  if (i < 0)				/* some error, end it */
     bb->inbuf = i;			/* forget the buffer */
   else if (bb->inbuf != 0)		/* trying to send line by line */
     bb->bufpos += i;			/* there is something left */
+  else
+    bb->bufpos = 0;			/* nothing left, reset head */
   return bb->inbuf;
 }
 
@@ -291,8 +306,18 @@ static ssize_t _ccfilter_x_send (connchain_i **ch, idx_t id, const char *str,
   if (bb->inbuf > 0)			/* there is something to send */
   {
     i = _ccfilter_x_push(ch, id, bb);
-    if (i != 0)				/* either not empty yet or error */
-      return ((i > 0) ? 0 : i);
+    if (i < 0)				/* either not empty yet or error */
+      return i;
+    else if (i > 0)
+    {
+      if (*sz > sizeof(bb->buf) - i - 2)
+	return 0;			/* no room for the line */
+      if (bb->bufpos > 0 && *sz > sizeof(bb->buf) - bb->bufpos - i - 2)
+      {
+	memmove(bb->buf, &bb->buf[bb->bufpos], i);
+	bb->bufpos = 0;
+      }
+    }
   }
   if (str == NULL)			/* got flush request */
     i = 0;
@@ -302,13 +327,13 @@ static ssize_t _ccfilter_x_send (connchain_i **ch, idx_t id, const char *str,
     return Connchain_Put (ch, id, str, sz); /* bounce test to next link */
   if (i > (ssize_t)sizeof(bb->buf) - 2)	/* line + CR/LF */
     i = sizeof(bb->buf) - 2;
-  memcpy (bb->buf, str, i);
+  memcpy (&bb->buf[bb->bufpos], str, i);
   DBG("connchain.c:_ccfilter_x_send: added to buffer +%zd", i);
   *sz -= i;
+  i += bb->inbuf;
   bb->buf[i] = '\r';			/* CR */
   bb->buf[i+1] = '\n';			/* LF */
   bb->inbuf = i + 2;
-  bb->bufpos = 0;
   _ccfilter_x_push(ch, id, bb);		/* try to send, ignoring errors now */
   return i;
 }
