@@ -1,5 +1,5 @@
 /*
- * Copyright (C) 1999-2014  Andrej N. Gritsenko <andrej@rep.kiev.ua>
+ * Copyright (C) 1999-2017  Andrej N. Gritsenko <andrej@rep.kiev.ua>
  *
  *     This program is free software; you can redistribute it and/or modify
  *     it under the terms of the GNU General Public License as published by
@@ -18,20 +18,28 @@
  * This file is part of FoxEye's source: common includes and definitions.
  */
 
+#ifdef _FOXEYE_H
+# error File foxeye.h can be included only once!
+#endif
+#define _FOXEYE_H 1
+
+#ifdef HAVE_CONFIG_H
+# include "../config.h"
+#endif
+
 /* strict POSIX.2 */
 /* #define _POSIX2_SOURCE 1 */
 #define _POSIX_C_SOURCE 200112L
 #define _XOPEN_SOURCE 700
+/* struct ucred is not available without _GNU_SOURCE so leave it set on Linux */
+#ifndef _LINUX
 #undef _GNU_SOURCE
+#endif
 
 #ifdef __sun
 # pragma error_messages (off, E_ARG_INCOMPATIBLE_WITH_ARG_L)
 #endif
 
-#ifdef HAVE_CONFIG_H
-# include "../config.h"
-#endif
-
 #ifdef HAVE_STDINT_H
 # include <stdint.h>
 #endif
@@ -39,20 +47,28 @@
 /* this is need for recursive thread under Linux */
 #define __USE_UNIX98 1
 
+#if TIME_WITH_SYS_TIME
+# include <sys/time.h>
+# include <time.h>
+#else
+# if HAVE_SYS_TIME_H
+#  include <sys/time.h>
+# else
+#  include <time.h>
+# endif
+#endif
+
 #include <stdio.h>
 #include <sys/types.h>
 #include <sys/stat.h>
 #include <stdarg.h>
 #include <pthread.h>
+#include <signal.h>
 
 #ifdef HAVE_THREAD_H
 # include <thread.h>
 #endif
 
-#ifdef _POSIX_SOURCE
-# include <time.h>
-#endif
-
 #if defined STDC_HEADERS || defined _LIBC
 # include <stdlib.h>
 #else
@@ -79,17 +95,6 @@ void free ();
 # include <unistd.h>
 #endif
 
-#if TIME_WITH_SYS_TIME
-# include <sys/time.h>
-# include <time.h>
-#else
-# if HAVE_SYS_TIME_H
-#  include <sys/time.h>
-# else
-#  include <time.h>
-# endif
-#endif
-
 #if HAVE_SYS_WAIT_H
 # include <sys/wait.h>
 #endif
@@ -185,6 +190,7 @@ void free ();
 #define HOSTMASKLEN	NAMEMAX+IDENTLEN+HOSTLEN+2 /* n!i@h or :p@h/n%f */
 #define RESTRICTEDCHARS	" ,\007\r\n" /* cannot be in names */
 #define MESSAGEMAX	1024	/* including '\0' at end */
+#define MBMESSAGEMAX	MB_LEN_MAX*MESSAGEMAX /* with NLS in mind */
 #define FORMATMAX	120	/* max lenght of format string */
 
 /* common Tcl/internal function prototypes */
@@ -213,7 +219,8 @@ typedef enum
   I_FINWAIT	= (1<<17),	/* from thread for dispatcher: terminate me */
   I_LOCKED	= (1<<18),	/* this interface temporary unavailable */
   I_AUTHPASS	= (1<<19),	/* client is authenticated */
-  I_DIED	= (1<<20)	/* this interface now died */	/* 0x100000 */
+  I_DIED	= (1<<20),	/* this interface now died */	/* 0x100000 */
+  I_SLEEPING	= (1<<21)	/* marked by dispatcher to wait */
 } iftype_t;
 
 /* logging/request flags */
@@ -249,9 +256,10 @@ typedef enum
   F_ASK		= (1<<24),	/*     required answer to that message */
   F_SIGNAL	= (1<<25),	/*     generate signal instead of request */
   F_END		= (1<<26),	/*     client going to down */
-  F_ECHO	= (1<<27),	/*     session only: echo on     --- session -*/
-  F_COLOR	= (1<<28),	/*     session only: enable color */
-  F_COLORCONV	= (1<<29)	/*     session only: convert color to ansi/mono */
+  F_RAW		= (1<<27),	/*     don't use charset conversions on this */
+  F_ECHO	= (1<<28),	/*     session only: echo on     --- session -*/
+  F_COLOR	= (1<<29),	/*     session only: enable color */
+  F_COLORCONV	= (1<<30)	/*     session only: convert color to ansi/mono */
 } flag_t;
 
 #define FLAG_T		"pmnucbseodijkw"
@@ -289,7 +297,8 @@ typedef enum
   S_STOP,			/* pause job on interface */
   S_CONTINUE,			/* continue stopped job */
   S_SHUTDOWN,			/* flush and terminate interface (quick) */
-  S_TERMINATE			/* terminate interface and release resources */
+  S_TERMINATE,			/* terminate interface and release resources */
+  S_WAKEUP			/* mark interace to run in any case */
 } __attribute__ ((packed)) ifsig_t;
 
 /* return values for _if_t::IFRequest */
@@ -314,6 +323,7 @@ struct _if_t
 #endif
   iftype_t ift;			/* flagged type of interface */
   int qsize;			/* queue FIFO size */
+  volatile sig_atomic_t marked;	/* marked to do NULL request */
 };
 
 typedef enum
@@ -362,6 +372,7 @@ typedef enum
 			/*'global AND channel' instead of default OR behavior */
 #define U_EQUAL	U_ALIAS	/* channel flag, for Check_Bintable() only, means */
 			/*all flags and mask are matched when they are equal */
+	/* U_NEGATE for Check_Bintable() has special meaning: to never match */
 
 #define U_NONE  (~(U_NEGATE | U_ALIAS))	/* for Add_Binding() */
 #define U_ALL	(~U_NEGATE)		/* global for Check_Bindtable() */
@@ -399,7 +410,8 @@ typedef enum
   A_REOP	= (1<<24),	/* server reop */
   A_TOPICLOCK	= (1<<25),	/* only ops may change topic */
   A_NOCOLOR	= (1<<26),	/* no color messages */
-  A_ASCIINICK	= (1<<27)	/* us-ascii nicks only */
+  A_ASCIINICK	= (1<<27),	/* us-ascii nicks only */
+  A_REGONLY	= (1<<28)	/* identified users only */
 } modeflag;
 
 #define A_MODEMASK	(A_MASKED-1)
@@ -477,6 +489,32 @@ typedef short idx_t;
 #define	IRCPAR_CHANMODES	"CHANMODES"	/* string, default= */
 #define	IRCPAR_CASEMAPPING	"CASEMAPPING"	/* string, default= */
 #define	IRCPAR_UMODES		"UMODES"	/* string, from version */
+#ifdef RUSNET_COMPILE
+
+/* module "ircd-rusnet" defines */
+
+#undef NO_DIRECT_VHOST		/* define to allow set +x only by services */
+#undef OPER_RCPAGE		/* define to allow opers to change user's charset */
+#define OPER_TLINE 1		/* define to allow oper to use KLINE, etc. locally */
+#undef LOCALOP_TLINE		/* define to allow local-ops to use KLINE, etc. */
+
+#define DEFAULT_TLINE_HOURS "8760" /* an year if timer was not set in KLINE, etc. */
+
+/* do NOT CHANGE next values, those are RusNet settings, changing them
+   will lead to the network desync! */
+
+#define NO_DEFAULT_VHOST 1	/* define to not set +x on connect by default */
+
+/* following settings are for check if client is really Rusnet Services */
+#define SERVICES_SERV "*.RusNet"
+#define SERVICES_IDENT "Service"
+#define SERVICES_HOST "RusNet"
+
+/* override settings of basic ircd to RusNet defines, cannot be above NICKLEN */
+#define RUSNET_NICKLEN		31
+
+#endif /* RUSNET_COMPILE */
+
 
 /* module "ircd" defines */
 
@@ -500,6 +538,8 @@ typedef short idx_t;
 
 #define TOPICWHOTIME 1		/* define to send RPL_TOPICWHOTIME to users */
 
+#define SEND_WHOIS_NOTICE 1	/* define to send NOTICE to oper on WHOIS on them */
+
   /* define to enable using these commands below by operators */
 #define IRCD_ENABLE_KILL 1
 #define IRCD_ENABLE_REHASH 1
@@ -522,6 +562,7 @@ typedef short idx_t;
 #define CHIDLEN		5	/* id len on ! channels */
 #define CHASETIMELIMIT	90	/* "nick delay" for RFC2813 servers */
 #define REOP_DELAY	5400	/* opless timelimit for +r channels */
+#define IRCD_TRUST_SERVER_NAME 1 /* define to 1 if servername may not match RFC1035 */
 
 #define IRCD_USES_ICONV 1	/* define to 1 if you want unicode IRCd */
 #define IRCD_STRICT_NAMES 1	/* define to 1 to strict names to CHARSET_8BIT */
@@ -562,8 +603,9 @@ typedef short idx_t;
 #define BINDING_TYPE_unload(func) static\
   void func(char *)
 #define BINDING_TYPE_login(func) static\
-  void func(char *, char *, char *, struct peer_t *,\
-		  char [SHORT_STRING], char **)
+  void func(char *, char *, const char *,\
+		  struct peer_t *, char [SHORT_STRING],\
+		  char **)
 #define BINDING_TYPE_connect(func) static\
   int func(const char *, char *)
 #define BINDING_TYPE_register(func) static\
@@ -583,6 +625,10 @@ typedef short idx_t;
   modeflag func(const char *, const char *,\
 		      const char *, const char **,\
 		      const char **, time_t *, short *)
+#define BINDING_TYPE_update_public(func) static\
+  int func(const char *, const char *, modeflag ,\
+	   const char *, const char *, const char *,\
+	   time_t )
 #define BINDING_TYPE_time_shift(func) static\
   void func(int )
 #define BINDING_TYPE_connchain_grow(func) static\
@@ -679,6 +725,8 @@ typedef short idx_t;
 		  char *, char *)
 #define BINDING_TYPE_irc_quit(func) static\
   int func(INTERFACE *, char *, size_t )
+#define BINDING_TYPE_ircd_capab(func) static\
+  void func (INTERFACE *, struct peer_t *, const char *)
 #define BINDING_TYPE_ircd_auth(func) static\
   int func(struct peer_t *, char *, char *, const char **,\
 	   modeflag *)
@@ -709,8 +757,9 @@ typedef short idx_t;
 #define BINDING_TYPE_ircd_check_message(func) static\
   int func(modeflag , modeflag , char *)
 #define BINDING_TYPE_ircd_check_modechange(func) static\
-  int func(INTERFACE *, modeflag , modeflag , int ,\
-	   modeflag , char *, modeflag , modeflag )
+  int func(INTERFACE *, modeflag , const char *, modeflag ,\
+	   int , modeflag , const char *, modeflag ,\
+	   modeflag )
 #define BINDING_TYPE_ircd_modechange(func) static\
   modeflag func(modeflag , modeflag ,\
 		      const char *, modeflag ,\
@@ -732,8 +781,8 @@ typedef short idx_t;
 #define BINDING_TYPE_ircd_stats_reply(func) static\
   void func(INTERFACE *, const char *, modeflag )
 #define BINDING_TYPE_ircd_collision(func) static\
-  char *func(INTERFACE *, char *, size_t , int ,\
-		   const char *, const char *)
+  int func(INTERFACE *, char *, size_t , const char *,\
+	   const char *)
 #define BINDING_TYPE_ircd_set_message_targets(func) static\
   int func (INTERFACE *, const char *, const char *,\
 	    int (*)(INTERFACE *, const char *,\
@@ -741,10 +790,23 @@ typedef short idx_t;
 			void **))
 #define BINDING_TYPE_ircd_whois(func) static\
   void func(INTERFACE *, const char *, modeflag ,\
-		  const char *, const char *, const char *,\
-		  modeflag )
+		  const char *, const char *,\
+		  const char *, modeflag )
 #define BINDING_TYPE_ircd_check_send(func) static\
   int func(INTERFACE *, struct peer_t *, modeflag , char *,\
 	   size_t )
 #define BINDING_TYPE_ircd_isupport(func) static\
   void func(char *, size_t )
+#define BINDING_TYPE_ircd_got_server(func) static\
+  void func (INTERFACE *, struct peer_t *, modeflag ,\
+		   unsigned short , const char *)
+#define BINDING_TYPE_ircd_lost_server(func) static\
+  void func (INTERFACE *, struct peer_t *)
+#define BINDING_TYPE_ircd_eob(func) static\
+  void func(INTERFACE *, struct peer_t *,\
+		  const char *, modeflag , int )
+#define BINDING_TYPE_ircd_drop_unknown(func) static\
+  void func(INTERFACE *, struct peer_t *, const char *,\
+		  const char *)
+#define BINDING_TYPE_ircd_server_handshake(func) static\
+  int func(INTERFACE *, struct peer_t *, const char *)
