@@ -891,8 +891,7 @@ int ircch_parse_modeline (IRC *net, CHANNEL *chan, LINK *origin, char *prefix,
 	  ircch_add_mask (list, "", 0, schr);
 	/* start enforcer if enforcing is on and ban is raised */
 	if ((gcf & U_AUTO) && chan->tid == -1 && mc == 'b' && mf == '+')
-	  chan->tid = NewTimer (chan->chi->ift, chan->chi->name, S_LOCAL,
-				ircch_enforcer_time, 0, 0, 0);
+	  chan->tid = Add_Timer (chan->chi, S_LOCAL, ircch_enforcer_time);
 	/* set matched exceptions while ban is set */
 	if (mc == 'b' &&		/* ignore +e and +I */
 	    !(net->features & L_NOEXEMPTS) &&
@@ -1511,7 +1510,7 @@ static int ssirc_kick (struct peer_t *who, INTERFACE *where, char *args)
     return 0;
   else if (!me)
   {
-    New_Request (who->iface, 0, _("No operator priveleges on %s."),
+    New_Request (who->iface, 0, _("No operator privileges on %s."),
 		 where->name);
     return -1;
   }
@@ -1575,7 +1574,7 @@ static int ssirc_kickban (struct peer_t *who, INTERFACE *where, char *args)
     return 0;
   else if (!me)
   {
-    New_Request (who->iface, 0, _("No operator priveleges on %s."),
+    New_Request (who->iface, 0, _("No operator privileges on %s."),
 		 where->name);
     return -1;
   }
@@ -1647,7 +1646,7 @@ static int _ssirc_setmode (struct peer_t *who, INTERFACE *where, char *args,
     return 0;
   else if (!me)
   {
-    New_Request (who->iface, 0, _("No operator priveleges on %s."),
+    New_Request (who->iface, 0, _("No operator privileges on %s."),
 		 where->name);
     return -1;
   }
@@ -1743,7 +1742,7 @@ static int ssirc_reset (struct peer_t *who, INTERFACE *where, char *args)
     return 0;
   else if (!me)
   {
-    New_Request (who->iface, 0, _("No operator priveleges on %s."),
+    New_Request (who->iface, 0, _("No operator privileges on %s."),
 		 where->name);
     return -1;
   }
@@ -1782,7 +1781,7 @@ static int ssirc_invite (struct peer_t *who, INTERFACE *where, char *args)
     return 0;
   else if ((ch->mode & A_INVITEONLY) && !me)
   {
-    New_Request (who->iface, 0, _("No operator priveleges on %s."),
+    New_Request (who->iface, 0, _("No operator privileges on %s."),
 		 where->name);
     return -1;
   }
