@@ -21,6 +21,7 @@
 /* these are used in structures below */
 #include <direct.h>
 #include <tree.h>
+#include <sheduler.h>
 
 	/* "modes" for internal usage */
 #define A_ISON		(1<<0)
@@ -65,6 +66,22 @@
 # define IRCD_ID_HISTORY 65536
 #endif
 
+/* Detailed explanation of structures relations:
+ * 1) each client or server, or phantom is present in names tree (IRCD)client
+ *    by lcnick;
+ * 2) each client or server, while registered, has associated LINK which is
+ *    present in closer server list:
+ *    - for local client it's (IRCD)users :: CLASS :: LINK...
+ *    - for local server it's (IRCD)servers :: LINK...
+ *    - for remote client or server it's (CLIENT)c.lients :: LINK...
+ * 3) each remote client has a backward reference to the server as (CLIENT)cs,
+ *    for remote server it's reference to the server itself;
+ * 4) each local client or server LINK is also present in associated peer_priv
+ *    structure, it's the same LINK which is referenced via (IRCD);
+ * 5) each local or remote server has associated token in a list:
+ *    - local list is (IRCD)token[]
+ *    - each peer list is (peer_priv)i.token[] */
+
 typedef struct CLASS CLASS;		/* internally defined in ircd.c */
 typedef struct CLIENT CLIENT;		/* defined below */
 typedef struct MEMBER MEMBER;
@@ -141,7 +158,8 @@ typedef struct LINK
 #define LINK_FL_GOT_ID 0x1
 #endif
 
-/* pcl is CLASS->glob thru all servers including us or NULL if it's server
+/* pcl is CLASS->glob thru all servers including us if it's user/service
+   or pointer to the previous server hop if it's server
    everything is valid only when via->s.state != P_DISCONNECTED
    via is valid only if it's local client or any server
    and if it's remote server then it's the shortest way to send message
@@ -267,6 +285,7 @@ struct peer_priv
   ACK *acks;		/* server: list of waiting acks */
 #endif
   pthread_t th;		/* thread ID if p.state == P_DISCONNECTED */
+  tid_t timer;		/* login or ping timeout timer id */
 };
 
 /*
@@ -387,7 +406,7 @@ ACK *ircd_check_ack (struct peer_priv *, CLIENT *, CHANNEL *);
 /* removes all acks from link */
 #define ircd_clear_acks(I,pp) while (pp->acks) ircd_drop_ack (I,pp)
 /* generates new id for new type commands from my clients or me */
-int ircd_new_id (void);
+int ircd_new_id (CLIENT *);
 /* if id isn't received yet then register it and return 1, else return 0 */
 int ircd_test_id (CLIENT *, int);
 #endif
@@ -416,7 +435,7 @@ void ircd_drop_channel (IRCD *, CHANNEL *);
 /* args: buf, umode, sizeof(buf); returns buf */
 char *ircd_make_umode (char *, modeflag, size_t);
 /* returns whochar for first appropriate mode */
-char ircd_mode2whochar (modeflag);
+char *ircd_mode2whochar (modeflag, char *, size_t);
 /* does backward conversion */
 modeflag ircd_whochar2mode(char);
 /* args: ircd, servername, nick, channame, mode char
@@ -428,10 +447,12 @@ modeflag ircd_char2umode(INTERFACE *, const char *, char, CLIENT *);
 
 	/* calls to channel.c from ircd.c */
 void ircd_channels_flush (IRCD *, char *, size_t);
-void ircd_burst_channels (INTERFACE *, NODE *);
+void ircd_burst_channels (struct peer_t *, IRCD *, modeflag);
 void ircd_channels_report (INTERFACE *);
 void ircd_channels_chreop (IRCD *, CLIENT *);
 void send_isupport (IRCD *, CLIENT *);
+bool ircd_check_modechange(INTERFACE *, modeflag, const char *, modeflag, int,
+			   modeflag, const char *, modeflag, modeflag);
 
 	/* calls to client.c from channel.c */
 int ircd_names_reply (CLIENT *, CLIENT *, CHANNEL *, int);
@@ -457,6 +478,9 @@ void ircd_queries_register (void);
 	/* messages.c bindings */
 void ircd_message_proto_start (void);
 void ircd_message_proto_end (void);
+	/* management.c bindings */
+void ircd_management_proto_start (void);
+void ircd_management_proto_end (void);
 
 	/* from queries.c */
 const char *ircd_version_flags;
