@@ -123,7 +123,6 @@ int ircd_names_reply (CLIENT *me, CLIENT *cl, CHANNEL *ch, int done)
   MEMBER *cm = ch->users, *here;
   size_t p, s, x;
   char buf[IRCMSGLEN-24]; /* :myname INUM xxxxxx 353 Nick = #chan :@some */
-  register char c;
 
   if (ch->mode & (A_QUIET | A_ANONYMOUS)) /* no listing */
     return done;
@@ -150,13 +149,11 @@ int ircd_names_reply (CLIENT *me, CLIENT *cl, CHANNEL *ch, int done)
       }
       s = 0;
     }
-    if (cm->mode & A_ADMIN)
-      c = '@';
-    else
-      c = ircd_mode2whochar (cm->mode);
-    buf[p+s] = c;
-    if (c)
+    if (cm->mode & A_ADMIN) {
+      buf[p+s] = '@';
       s++;
+    } else
+      s += strlen(ircd_mode2whochar (cm->mode, &buf[p+s], sizeof(buf)-p-s));
     s += snprintf (&buf[p+s], x - p - s, "%s ", cm->who->nick);
   } while ((cm = cm->prevnick));
   if (s)
@@ -514,14 +511,14 @@ static inline int _ircd_query_lusers (IRCD *ircd, CLIENT *cl, struct peer_priv *
 	  ircd->token[0]->x.a.uc, lu + op);
     ircd->token[0]->x.a.uc = lu + op;
   }
-  if (ls != 0)
+  if (ll != 0)
   {
-    ERROR("ircd:/lusers: found servers in clients on ME: %u != 0", ls);
-    ls = 0;
+    ERROR("ircd:/lusers: found servers in clients on ME: %u != 0", ll);
+    ll = 0;
   }
   for (l = ircd->servers; l; l = l->prev)
     if (CLIENT_IS_SERVER(l->cl))
-      ls++;
+      ll++;
     else
       ERROR("ircd:/lusers: client %s in local servers list isn't a server",
 	    l->cl->nick);
@@ -542,7 +539,7 @@ static inline int _ircd_query_lusers (IRCD *ircd, CLIENT *cl, struct peer_priv *
       }
 #define COUNT_USERS_CHECK_STATE 
 #if IRCD_MULTICONNECT
-# define COUNT_USERS_CHECK_PATH if (l->cl->via != ircd->token[i]->via) continue; else
+# define COUNT_USERS_CHECK_PATH if (l->cl->via != NULL && l->cl->via != ircd->token[i]->via) continue; else
 #else
 # define COUNT_USERS_CHECK_PATH 
 #endif
@@ -582,7 +579,7 @@ static inline int _ircd_query_lusers (IRCD *ircd, CLIENT *cl, struct peer_priv *
   snprintf (buff, sizeof(buff), "%u", ircd->lu);
   ircd_do_unumeric (cl, RPL_LOCALUSERS, cl, (unsigned short)lu, buff);
   snprintf (buff, sizeof(buff), "%u", ircd->gu);
-  return ircd_do_unumeric (cl, RPL_GLOBALUSERS, cl, (unsigned short)gu, buff);
+  return ircd_do_unumeric (cl, RPL_GLOBALUSERS, cl, (unsigned short)(gu + op), buff);
 }
 
 BINDING_TYPE_ircd_client_cmd(ircd_lusers_cb);
@@ -632,6 +629,9 @@ const char *ircd_version_flags =
 #ifdef IRCD_ENABLE_USERS
 "U"
 #endif
+#ifdef SEND_WHOIS_NOTICE
+"W"
+#endif
 #ifdef ENABLE_IPV6
 "6"
 #endif
@@ -727,7 +727,7 @@ static void _ircd_tell_links(CLIENT *cl, CLIENT *server, char *where,
   for (tgt = server->c.lients; tgt; tgt = tgt->prev)
     if (CLIENT_IS_SERVER(tgt->cl) &&
 #if IRCD_MULTICONNECT
-	(tgt->cl != server) && (tgt->cl->via == via) && /* ignore backlink */
+	(tgt->cl->pcl == server) && /* ignore backlink */
 #endif
 	simple_match (smask, tgt->cl->lcnick) >= 0)
       _ircd_tell_links(cl, tgt->cl, server->lcnick, smask, via);
@@ -1018,23 +1018,13 @@ static void _ircd_do_whois (IRCD *ircd, CLIENT *cl, CLIENT *tgt, CLIENT *me)
 {
   MEMBER *m;
   size_t ptr;
-  char mc;
   char buf[IRCMSGLEN];
-  register struct binding_t *b;
+  register struct binding_t *b = NULL;
+#define static register
+  BINDING_TYPE_ircd_whois ((*f));
+#undef static
 
   ircd_do_unumeric (cl, RPL_WHOISUSER, tgt, 0, tgt->fname);
-  if (!CLIENT_IS_REMOTE (tgt)) {
-    ircd_do_unumeric (cl, RPL_WHOISSERVER, me, 0, tgt->nick);
-    //TODO: ifdef WHOIS_SIGNON_TIME => tgt->via->started
-    snprintf (buf, sizeof(buf), "%u", (unsigned int)(Time - tgt->via->noidle));
-    ircd_do_unumeric (cl, RPL_WHOISIDLE, tgt, 0, buf);
-  } else
-    ircd_do_unumeric (cl, RPL_WHOISSERVER, tgt->cs, 0, tgt->nick);
-  if (tgt->umode & (A_OP | A_HALFOP))
-    //TODO: notify target about whois on them
-    ircd_do_unumeric (cl, RPL_WHOISOPERATOR, tgt, 0, NULL);
-  if (tgt->umode & A_AWAY) /* FIXME: make a message "Use /WHOIS %s %s" ? */
-    ircd_do_unumeric (cl, RPL_AWAY, tgt, 0, tgt->away[0] ? tgt->away : "Gone");
   ptr = 0;
   for (m = tgt->c.hannels; m; m = m->prevchan)
   {
@@ -1046,10 +1036,6 @@ static void _ircd_do_whois (IRCD *ircd, CLIENT *cl, CLIENT *tgt, CLIENT *me)
 	if (!_ircd_is_on_channel (cl, m->chan))
 	  continue;
     }
-    if (m->mode & A_ADMIN)
-      mc = '@';
-    else
-      mc = ircd_mode2whochar (m->mode);
     if (ptr + strlen (m->chan->name) > sizeof(buf) - 3) /* reserved for " +" */
     {
       ircd_do_unumeric (cl, RPL_WHOISCHANNELS, tgt, 0, buf);
@@ -1057,12 +1043,32 @@ static void _ircd_do_whois (IRCD *ircd, CLIENT *cl, CLIENT *tgt, CLIENT *me)
     }
     if (ptr)
       buf[ptr++] = ' ';
-    if (mc)
-      buf[ptr++] = mc;
+    if (m->mode & A_ADMIN)
+      buf[ptr++] = '@';
+    else
+      ptr += strlen(ircd_mode2whochar(m->mode, &buf[ptr], sizeof(buf) - ptr));
     ptr += strfcpy (&buf[ptr], m->chan->name, sizeof(buf) - ptr);
   }
   if (ptr)
     ircd_do_unumeric (cl, RPL_WHOISCHANNELS, tgt, 0, buf);
+  if (!CLIENT_IS_REMOTE (tgt))
+    ircd_do_unumeric (cl, RPL_WHOISSERVER, me, 0, tgt->nick);
+  else
+    ircd_do_unumeric (cl, RPL_WHOISSERVER, tgt->cs, 0, tgt->nick);
+  if (tgt->umode & (A_OP | A_HALFOP))
+  {
+    ircd_do_unumeric (cl, RPL_WHOISOPERATOR, tgt, 0, NULL);
+#ifdef SEND_WHOIS_NOTICE
+    /* notify target about whois on them, unless it's on itself */
+    if (tgt != cl)
+      New_Request (tgt->cs->via->p.iface, 0,
+		   ":%s NOTICE %s :WHOIS on YOU requested by %s (%s@%s) [%s]",
+		   me->lcnick, tgt->nick, cl->nick, cl->user, cl->host,
+		   CLIENT_IS_LOCAL(cl) ? me->lcnick : cl->cs->nick);
+#endif
+  }
+  if (tgt->umode & A_AWAY) /* FIXME: make a message "Use /WHOIS %s %s" ? */
+    ircd_do_unumeric (cl, RPL_AWAY, tgt, 0, tgt->away[0] ? tgt->away : "Gone");
 #if IRCD_USES_ICONV
   if (!CLIENT_IS_REMOTE (tgt))
     ircd_do_unumeric (cl, RPL_WHOISCHARSET, tgt, 0,
@@ -1071,9 +1077,22 @@ static void _ircd_do_whois (IRCD *ircd, CLIENT *cl, CLIENT *tgt, CLIENT *me)
   /* special support for SSL client connections */
   if (tgt->umode & A_SSL)
     ircd_do_unumeric (cl, RPL_WHOISSECURE, tgt, 0, NULL);
-  while ((b = Check_Bindtable (BTIrcdWhois, tgt->nick, U_ALL, U_ANYCH, NULL)))
+  while ((b = Check_Bindtable (BTIrcdWhois, tgt->nick, U_ALL, U_ANYCH, b)))
     if (b->name == NULL)
-      b->func (ircd, cl->nick, cl->umode, tgt->nick, tgt->host, tgt->vhost, tgt->umode);
+    {
+      f = (void(*)())b->func;
+      f(ircd->iface, cl->nick, cl->umode, tgt->nick, tgt->host, tgt->vhost, tgt->umode);
+    }
+  /* binding might send some message, deliver it to user before RPL_ENDOFWHOIS */
+  Set_Iface(ircd->iface);
+  while(Get_Request());
+  Unset_Iface();
+  /*  */
+  if (!CLIENT_IS_REMOTE (tgt)) {
+    //TODO: ifdef WHOIS_SIGNON_TIME => tgt->via->started
+    snprintf (buf, sizeof(buf), "%u", (unsigned int)(Time - tgt->via->noidle));
+    ircd_do_unumeric (cl, RPL_WHOISIDLE, tgt, 0, buf);
+  }
 }
 
 static inline int _ircd_query_whois (IRCD *ircd, CLIENT *cl, struct peer_priv *via,
@@ -1463,6 +1482,7 @@ static void _icchg_ww(INTERFACE *srv, const char *from, const char *lcnick,
     return;
   if (nick == NULL)
     return;
+  dprint(5, "ircd:queries.c:_icchg_ww: %s (%s@%s)", nick, user, host);
   /* check if we are at end of array and want to grow it */
   if (IrcdWhowasPtr == IrcdWhowasSize && left < (INT_MAX / 2 - SOCKETMAX) &&
       left<<1 > (unsigned int)IrcdWhowasSize) {
@@ -1762,7 +1782,7 @@ void ircd_queries_proto_start (void)
   Add_Binding ("ircd-client-cmd", "stats", 0, 0, &ircd_stats_cb, NULL);
   Add_Binding ("ircd-client-cmd", "links", 0, 0, &ircd_links_cb, NULL);
   Add_Binding ("ircd-client-cmd", "time", 0, 0, &ircd_time_cb, NULL);
-  Add_Binding ("ircd-client-cmd", "connect", 0, 0, &ircd_connect_cb, NULL);
+  Add_Binding ("ircd-client-cmd", "connect", 0, U_HALFOP, &ircd_connect_cb, NULL);
   Add_Binding ("ircd-client-cmd", "trace", 0, 0, &ircd_trace_cb, NULL);
   Add_Binding ("ircd-client-cmd", "admin", 0, 0, &ircd_admin_cb, NULL);
   Add_Binding ("ircd-client-cmd", "info", 0, 0, &ircd_info_cb, NULL);
