@@ -62,7 +62,7 @@ void ircd_add_ack (struct peer_priv *link, CLIENT *who, CHANNEL *where)
   ack->contrary = 0;
   if (where != NULL && where != CHANNEL0)
     where->on_ack++;
-  dprint(3, "ircd:serverc.s: new ack: who=%p where=%p", who, where);
+  dprint(2, "ircd:serverc.s: new ack: who=%p where=%p", who, where);
 }
 
 /* removes first ack from link */
@@ -70,7 +70,7 @@ void ircd_drop_ack (IRCD *ircd, struct peer_priv *link)
 {
   register ACK *ack = link->acks;
 
-  dprint(3, "ircd:serverc.s: del ack: who=%p where=%p", ack->who, ack->where);
+  dprint(2, "ircd:serverc.s: del ack: who=%p where=%p", ack->who, ack->where);
   link->acks = ack->next;
   if (ack->who) {
     ack->who->on_ack--;			/* unlock one more */
@@ -98,9 +98,7 @@ ACK *ircd_check_ack (struct peer_priv *link, CLIENT *who, CHANNEL *where)
         possible = ack;
     } else if (ack->who == who && ack->where == where)
       return ack;
-    else
-      return possible;
-  return NULL;
+  return possible;
 }
 
 static ACK *ircd_find_ack(struct peer_priv *link, const char *who, const char *where)
@@ -109,7 +107,8 @@ static ACK *ircd_find_ack(struct peer_priv *link, const char *who, const char *w
 
   for (ack = link->acks; ack; ack = ack->next)
   {
-    if (where == NULL && ack->where == NULL && strcmp(who, ack->who->nick) == 0) {
+    if (where == NULL && ack->where == NULL &&
+	(ack->who == NULL || strcmp(who, ack->who->nick)) == 0) {
       if (ack->contrary)
 	possible = ack;
       else
@@ -134,16 +133,6 @@ static int localid = 0;
 
 #define IRCD_ID_MAXVAL 0x7fffffff /* INT32_MAX - for 32bit arch compatibility */
 
-/* generate new id for own messages */
-int ircd_new_id(void)
-{
-  if (localid == IRCD_ID_MAXVAL)
-    localid = 0;
-  else
-    localid++;
-  return (localid);
-}
-
 #define ID_MAP_MASK (IRCD_ID_HISTORY-1)
 
 /* bit manipulation functions; borrowed from SLURM sources */
@@ -275,6 +264,27 @@ int ircd_test_id(CLIENT *cl, int id)
   cl->last_id = id;
   return (1);
 }
+
+/* generate new id for own messages */
+int ircd_new_id(CLIENT *cl)
+{
+  if (cl)
+  {
+    if (!CLIENT_IS_SERVER(cl))
+      return (-1);
+    if (cl->last_id == IRCD_ID_MAXVAL)
+      cl->last_id = 0;
+    else
+      cl->last_id++;
+    bit_set((bitstr_t *)cl->id_cache, (cl->last_id & ID_MAP_MASK));
+    return (cl->last_id);
+  }
+  if (localid == IRCD_ID_MAXVAL)
+    localid = 0;
+  else
+    localid++;
+  return (localid);
+}
 #endif
 
 
@@ -334,19 +344,18 @@ static int ircd_squit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
     return ircd_recover_done(pp, "SQUIT need more parameters");
   }
   tgt = ircd_find_client(argv[0], pp);	/* in case of backfired it's NULL */
-  if (tgt == NULL || !CLIENT_IS_SERVER(tgt))
+  if (tgt == NULL || tgt->hold_upto || !CLIENT_IS_SERVER(tgt)) {
 #if IRCD_MULTICONNECT
-  {
-    if (tgt == NULL && (pp->link->cl->umode & A_MULTI)) {
-      New_Request(peer->iface, 0, "ACK SQUIT %s", argv[0]);
+    if ((tgt == NULL || tgt->hold_upto) && (pp->link->cl->umode & A_MULTI)) {
+      cl = _ircd_find_client_lc((IRCD *)srv->data, lcsender);
+      if (CLIENT_IS_SERVER(cl))
+	New_Request(peer->iface, 0, "ACK SQUIT %s", argv[0]);
       return (1);
       //TODO: log duplicate?
     } else
 #endif
       return ircd_recover_done(pp, "No such server");
-#if IRCD_MULTICONNECT
   }
-#endif
   cl = _ircd_find_client_lc((IRCD *)srv->data, lcsender);
   if (CLIENT_IS_SERVER(cl)) {
 #if IRCD_MULTICONNECT
@@ -362,11 +371,13 @@ static int ircd_squit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
       ack = ircd_check_ack(pp, tgt, NULL);
       if (ack != NULL) {
 	ack->contrary = 1;
-	return (1); /* ignore the message */
       }
-#endif
+      /* at this point we might have cleaned it as a backlink */
+      return (1); /* ignore the message */
+#else
       ircd_do_squit(pp->link, pp, "Invalid SQUIT message");
       return 0;				/* kill our peer instead */
+#endif
     }
     ircd_do_squit(l, pp, argv[1]);
   } else if (cl->hold_upto) {
@@ -377,8 +388,9 @@ static int ircd_squit_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
       ircd_sendto_wallops((IRCD *)srv->data, NULL, me, "SQUIT %s from %s: %s",
 			  argv[0], cl->nick, argv[1]);
       ircd_do_squit(tgt->via->link, NULL, argv[1]); /* do job */
-    } else				/* or else forward it to it's links */
-      ircd_sendto_remote(tgt, ":%s SQUIT %s :%s", cl->nick, argv[0], argv[1]);
+    } else {				/* or else forward it to its link */
+      ircd_sendto_one(tgt, ":%s SQUIT %s :%s", cl->nick, argv[0], argv[1]);
+    }
   }
   return 1;
 }
@@ -498,9 +510,9 @@ static int ircd_join_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
       else
 	m += ptr;
     } else
-#if !IRCD_MULTICONNECT
+#if IRCD_MULTICONNECT
     /* for multiconnected server don't set error */
-    if (!(pp->link->cl->umode & A_MULTI))
+    if (!(((struct peer_priv *)peer->iface->data)->link->cl->umode & A_MULTI))
 #endif
       err = 1;
     if (*c == ',')
@@ -526,9 +538,9 @@ static int ircd_njoin(INTERFACE *srv, struct peer_t *peer, unsigned short token,
   MEMBER *tgt;
   register MEMBER *t;
   CLIENT *cl;
-  const char *c;
+  const char *c, *n;
   struct peer_priv *pp = peer->iface->data; /* it's really peer */
-  size_t mptr;
+  size_t mptr, mflen;
   modeflag mf;
   register modeflag amf;
   char clname[MB_LEN_MAX*NICKLEN+1];
@@ -551,20 +563,19 @@ static int ircd_njoin(INTERFACE *srv, struct peer_t *peer, unsigned short token,
   mptr = 0;
   c = argv[1];
   while (*c) {
-    if (mptr < sizeof(msg)-1) {
-      if (mptr == 0)
-	msg[0] = *c;
-      else
-	msg[mptr+1] = *c;
-    }
+    n = c;
     if (c[0] == '@' && c[1] == '@') {
       mf = A_ADMIN;
-      c += 2;
+      mflen = 2;
     } else {
-      mf = ircd_whochar2mode(*c);
-      if (mf != 0)
-	c++;
+      mf = 0;
+      mflen = 0;
+      while ((amf = ircd_whochar2mode(c[mflen])) != 0) {
+	mf |= amf;
+	mflen++;
+      }
     }
+    c += mflen;
     for (i = 0; i < sizeof(clname)-1 && *c != '\0' && *c != ','; c++)
       clname[i++] = *c;
     while (*c != '\0' && *c++ != ',');	/* skip ',' if any */
@@ -593,14 +604,12 @@ static int ircd_njoin(INTERFACE *srv, struct peer_t *peer, unsigned short token,
       t = ircd_add_to_channel(srv->data, pp, tgt->chan, cl, mf);
     if (t == NULL)		/* not approved to add */
       continue; //TODO: make debug message?
-    if (mptr + strlen(clname) >= sizeof(msg) - 4) /* it's impossible! ERROR? */
+    if (mptr + mflen + strlen(clname) >= sizeof(msg) - 3) /* it's impossible! ERROR? */
       continue;
     if (mptr != 0)
       msg[mptr++] = ',';
-    if (mf != 0)
-      mptr++;			/* it already took the char, see above */
-    if (mf & A_ADMIN)
-      msg[mptr++] = '@';
+    for (i = 0; i < mflen; i++)
+      msg[mptr++] = n[i];
     mptr += strfcpy(&msg[mptr], clname, sizeof(msg) - mptr);
   }
   if (tgt) {			/* do broadcast to neighbours */
@@ -658,11 +667,11 @@ static int ircd_part_sb(INTERFACE *srv, struct peer_t *peer, unsigned short toke
     *t = '\0';
     memb = ircd_find_member ((IRCD *)srv->data, chname, NULL);
     if (memb == NOSUCHCHANNEL) {
-#if !IRCD_MULTICONNECT
+#if IRCD_MULTICONNECT
       /* for multiconnected server don't set error */
       if (pp->link->cl->umode & A_MULTI) {
-	DBG("ircd:got PART from %s for %s on nonexistent channel %s",
-	    peer->dname, sender, chname);
+	dprint(4, "ircd:got PART from %s for %s on nonexistent channel %s",
+	       peer->dname, sender, chname);
 	continue;
       }
 #endif
@@ -722,12 +731,20 @@ static int ircd_topic_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
   CLIENT *cl;
   MEMBER *memb;
   struct peer_priv *pp = peer->iface->data; /* it's really peer */
+  int id = -1;
 
   /* check number of parameters and if channel isn't local one */
   if (argc != 2) {
     ERROR("ircd:got TOPIC from %s with %d(<2) parameters", peer->dname, argc);
     return ircd_recover_done(pp, "Invalid number of parameters");
   }
+#if IRCD_MULTICONNECT
+  if (pp->link->cl->umode & A_MULTI)
+  {
+    ERROR("ircd:illegal TOPIC command via %s", peer->dname);
+    return ircd_recover_done(pp, "illegal TOPIC command");
+  }
+#endif
   cl = _ircd_find_client_lc((IRCD *)srv->data, lcsender);
   memb = ircd_find_member ((IRCD *)srv->data, argv[0], NULL);
   /* we should never get TOPIC from A_MULTI link so no reason to check acks */
@@ -736,8 +753,11 @@ static int ircd_topic_sb(INTERFACE *srv, struct peer_t *peer, unsigned short tok
 	  peer->dname, sender, argv[0]);
     return ircd_recover_done(pp, "TOPIC for nonexistent channel");
   }
-  return _ircd_do_stopic((IRCD *)srv->data, peer->dname, sender, pp, 0, -1, cl,
-			 memb->chan, argv[1]);
+#if IRCD_MULTICONNECT
+  id = ircd_new_id(cl->cs);
+#endif
+  return _ircd_do_stopic((IRCD *)srv->data, peer->dname, sender, pp, token, id,
+			 cl, memb->chan, argv[1]);
 }
 
 BINDING_TYPE_ircd_server_cmd(ircd_invite_sb);
@@ -779,7 +799,7 @@ static int ircd_invite_sb(INTERFACE *srv, struct peer_t *peer, unsigned short to
       ircd_add_invited (tgt, me->chan);
   }
   if (CLIENT_IS_REMOTE(tgt))
-    ircd_sendto_one (tgt, ":%s INVITE %s %s", sender, argv[0], argv[1]);
+    ircd_sendto_remote (tgt, cl, pp, ":%s INVITE %s %s", sender, argv[0], argv[1]);
   else
     ircd_sendto_one (tgt, ":%s!%s@%s INVITE %s %s", sender, cl->user, cl->vhost,
 		     argv[0], argv[1]);
@@ -1041,31 +1061,19 @@ static int _ircd_do_stopic(IRCD *ircd, const char *via, const char *sender,
 #ifdef USE_SERVICES
     ircd_sendto_services_mark_nick(ircd, SERVICE_WANT_TOPIC);
 #endif
-#if IRCD_MULTICONNECT
-    if (id >= 0) {
-      ircd_sendto_servers_mask_old(ircd, pp, cmask, ":%s TOPIC %s :%s", sender,
-				   ch->name, ch->topic);
-      ircd_sendto_servers_mask_new(ircd, pp, cmask, ":%s ITOPIC %d %s :%s",
-				   sender, id, ch->name, ch->topic);
-    } else
-#endif
-      ircd_sendto_servers_mask(ircd, pp, cmask, ":%s TOPIC %s :%s", sender,
-			       ch->name, ch->topic);
+    ircd_sendto_servers_mask_old(ircd, pp, cmask, ":%s TOPIC %s :%s", sender,
+				 ch->name, ch->topic);
+    ircd_sendto_servers_mask_new(ircd, pp, cmask, ":%s ITOPIC %d %s :%s",
+				 sender, id, ch->name, ch->topic);
     return 1;
   }
 #ifdef USE_SERVICES
   ircd_sendto_services_mark_nick(ircd, SERVICE_WANT_TOPIC);
 #endif
-#if IRCD_MULTICONNECT
-  if (id >= 0) {
-    ircd_sendto_servers_old(ircd, pp, ":%s TOPIC %s :%s",
-			    sender, ch->name, ch->topic);
-    ircd_sendto_servers_new(ircd, pp, ":%s ITOPIC %d %s :%s",
-			    sender, id, ch->name, ch->topic);
-  } else
-#endif
-    ircd_sendto_servers_all(ircd, pp, ":%s TOPIC %s :%s",
-			    sender, ch->name, ch->topic);
+  ircd_sendto_servers_old(ircd, pp, ":%s TOPIC %s :%s",
+			  sender, ch->name, ch->topic);
+  ircd_sendto_servers_new(ircd, pp, ":%s ITOPIC %d %s :%s",
+			  sender, id, ch->name, ch->topic);
   return 1;
 }
 
